
# 	***********************************************************	
# 	Using transfer entropy to measure information flows 
# 	between financial markets
#	Thomas Dimpfl and Franziska J. Peter
#	Studies in Nonlinear Dynamics and Econometrics
#	-----------------------------------------------------------	
# 	Calculation of Transfer Entropy
#	last modified 13 April 2012     
# 	***********************************************************	

# shell command: nohup R CMD BATCH cds_entropie_main.R &

library(multicore)
library(chron)



# load functions
source("entropie_procedures.R")

# load list of ticker symbols 
ind <- "AZ"

# define physical path to data
datapath <- "/home/dimpflth/public/research/dhp2010/data/Bonds_CDS_merged/"

# define physical path where results are to be stored
respath <- "/home/dimpflth/public/research/dhp2010/results/"



CALL_TE_CDS <- function(x,
                   y,
                   lx=2,
                   ly=2,
                   nrep=10, 
                   Konst=FALSE,
                   Shuffle=TRUE,
                   Klassen=3,
		   Quant=c(5,95),
                   nBoot=9999){


    # Kodierung der Prozesse
    #-------------------------

    x <- KodStp(x, Quant=Quant, Klassen=Klassen)  #$Kodierte_Stichprobe
    y <- KodStp(y, Quant=Quant, Klassen=Klassen)  #$Kodierte_Stichprobe
    
    # Berechnung der ersten Transfer Entropie
    #-----------------------------------------
    
    #-------------
    #Ohne Shuffle
    #-------------
    
    # Lead=X
    #---------
    
    TransferEntropie_X     <- TransEnt(X=x,Y=y,lX=lx,lY=ly)$TransEnt 
    
    # Lead=Y
    #---------
    
    TransferEntropie_Y     <- TransEnt(X=y,Y=x,lX=ly,lY=lx)$TransEnt 
    
    #-------------
    #Mit Shuffle
    #-------------
    
    
    
    KonstanteX <- Shuffled_TE(nrep=nrep, 
                            Diff=FALSE,
                            X = x,
                            lX = lx, 
                            Y = y,
                            lY = ly)
    
    
    KonstanteY <- Shuffled_TE(nrep=nrep, 
                            Diff=FALSE,
                            X = y,
                            lX = ly, 
                            Y = x,
                            lY = lx)
    
    
    # Lead=X
    #---------
    
    S_TransferEntropie_X  <-   TransferEntropie_X - KonstanteX
    
    # Lead=Y
    #---------
    
    S_TransferEntropie_Y   <-  TransferEntropie_Y - KonstanteY
    

    
   #----------------------Bootstrap-----------------------------#
    
       
    Boot   <-  replicate(nBoot, TransBoot_H0(X = x, 
					  Y = y,
                                          Shuffle = TRUE,
                                          lx = lx,
                                          ly = ly,
                                          Konst= TRUE,
                                          KonstX = 0,
                                          KonstY = 0,
                                          nrep=nrep))

    # Zusammenlegung der Stichprobe  
    #------------------------------

    Koll <- KollapseStp(X=x,Y=y)
    
    KX   <- Koll$KollStich

    # Berechnung der Wahrscheinlichkeiten
    #-------------------------------------
   
     
    Boot2   <-  replicate(nBoot, TransBoot_H1(X = KX, 
                                          AusprTab=Koll$AusprTab,
                                          Shuffle    = Shuffle,
                                          lx = lx,
                                          ly = ly,
                                          Konst=TRUE,
                                          KonstX=mean(Boot["DTEX",]),
                                          KonstY=mean(Boot["DTEY",]),
                                          nrep=10))
    
    
 
   out <- list(TransferEntropie_X   = TransferEntropie_X ,
               TransferEntropie_Y   = TransferEntropie_Y,
               S_TransferEntropie_X = S_TransferEntropie_X,
               S_TransferEntropie_Y = S_TransferEntropie_Y,
               Bootstrap_H0 = Boot,
	       Bootstrap_H1 = Boot2)
   
   
   return(out)
   
   
}




forall <- function(lags,ind,quant,dest_folder,crisis=0){

      for (i in 1:length(ind)){
      # for (i in 1:2){
	print(paste(ind[i],"started",sep=" "))
	indat <- dget(paste(datapath,ind[i],".RData",sep=""))

	indat <- indat[!is.na(indat[,"CDS"]),]

      #   restrict periods
	if (crisis=="before"){
	indat <- indat[dates(as.character(indat[,1]))<"07/01/2007",]
	}
	if (crisis=="crisis"){
	indat <- indat[dates(as.character(indat[,1]))>"06/30/2007",]
	indat <- indat[dates(as.character(indat[,1]))<"10/01/2009",]
	}
	if (crisis=="after"){
	indat <- indat[dates(as.character(indat[,1]))>"09/30/2009",]
	}

	spread <- as.numeric(as.character(indat[2:dim(indat)[1],"Spread"])) -as.numeric(as.character(indat[1:(dim(indat)[1]-1),"Spread"]))
	cds    <- as.numeric(as.character(indat[2:dim(indat)[1],"CDS"])) - as.numeric(as.character(indat[1:(dim(indat)[1]-1),"CDS"])) 
# 	print(cor(cbind(spread,cds)))



	Entropy <- CALL_TE_CDS (x=cds,
			y=spread,
			lx=lags[lags[,2]==as.character(ind[i]),1],
			ly=lags[lags[,2]==as.character(ind[i]),1],
			nrep=499, 
			Konst=FALSE,
			Shuffle=TRUE,
			Klassen=3,
			Quant = quant,
			nBoot=999)

	dput(Entropy,paste(respath,dest_folder,"CDS_TE_",ind[i],sep=""))
	cat("finished",ind[i],"\n")
      }

      print("end first loop")

      for (i in 1:length(ind)){
      # for (i in 1:18){

	TE <- dget(paste(respath,dest_folder,"CDS_TE_",ind[i],sep=""))

# 	TE$Bootstrap_H0["DTEX",]
# 	TE$Bootstrap_H0["DTEY",]

	tex <- format(TE$TransferEntropie_X,digits=5,scientific=FALSE)
	tey <- format(TE$TransferEntropie_Y,digits=5,scientific=FALSE)
	stex <- format(TE$S_TransferEntropie_X,digits=5,scientific=FALSE)
	stey <- format(TE$S_TransferEntropie_Y,digits=5,scientific=FALSE)
	sx <- format(TE$TransferEntropie_X-TE$S_TransferEntropie_X,digits=5,scientific=FALSE)
	sy <- format(TE$TransferEntropie_Y-TE$S_TransferEntropie_Y,digits=5,scientific=FALSE)
      # format(x,digits=5,scientific=FALSE)

	ndi <- TE$TransferEntropie_X - mean(TE$Bootstrap_H0["DTEX",]) - (TE$TransferEntropie_Y - 		    mean(TE$Bootstrap_H0["DTEY",]))

	qx <- format(quantile(TE$Bootstrap_H0["DTEX",],c(90,95,99)/100),digits=5,scientific=FALSE)
	qy <- format(quantile(TE$Bootstrap_H0["DTEY",],c(90,95,99)/100),digits=5,scientific=FALSE)

	sex <- format(var(TE$Bootstrap_H0["DTEX",]),digits=5,scientific=FALSE)
	sey <- format(var(TE$Bootstrap_H0["DTEY",]),digits=5,scientific=FALSE)

	te <- cbind(TE$Bootstrap_H0["DTEX",], TE$Bootstrap_H0["DTEY",])
	te[te[,1] <0,1]<-0
	te[te[,2] <0,2]<-0

	ndi_boot <- TE$Bootstrap_H1["DTEX",] - mean(TE$Bootstrap_H0["DTEX",]) - (TE$Bootstrap_H1["DTEY",] - mean(TE$Bootstrap_H0["DTEY",]))
      #   m_ndi <- mean(ndi_boot)
      #   ndi_boot <- ndi_boot - m_ndi
      #   ndi_boot[ndi_boot == "NaN"] = 0

	qndi <- format(quantile(ndi_boot,c(0.5,2.5,5,95,97.5,99.5)/100),digits=5,scientific=FALSE)
	se_ndi <- sd(ndi_boot)

	m_tex <- format(mean(TE$Bootstrap_H0["DTEX",]))
	m_tey <- format(mean(TE$Bootstrap_H0["DTEY",]))

	x_mboot <- format(TE$TransferEntropie_X - mean(TE$Bootstrap_H0["DTEX",]))
	y_mboot <- format(TE$TransferEntropie_Y - mean(TE$Bootstrap_H0["DTEY",]))

	x_h1_bootmean <- format(mean(TE$Bootstrap_H1["DTEX",]))
	y_h1_bootmean <- format(mean(TE$Bootstrap_H1["DTEY",]))

	qx_h1 <- format(quantile(TE$Bootstrap_H1["DTEX",],c(5,95)/100),digits=5,scientific=FALSE)
	qy_h1 <- format(quantile(TE$Bootstrap_H1["DTEY",],c(5,95)/100),digits=5,scientific=FALSE)

	sdx_h1 <- format(sd(TE$Bootstrap_H1["DTEX",]),digits=5,scientific=FALSE)
	sdy_h1 <- format(sd(TE$Bootstrap_H1["DTEY",]),digits=5,scientific=FALSE)

	tmp1 <- rbind(
	  c(quantile(TE$Bootstrap_H1["DTEX",],5/100),quantile(TE$Bootstrap_H1["DTEY",],5/100)),
	  c(TE$TransferEntropie_X - mean(TE$Bootstrap_H0["DTEX",]),TE$TransferEntropie_Y - mean(TE$Bootstrap_H0["DTEY",])),
	  c(quantile(TE$Bootstrap_H1["DTEX",],95/100),quantile(TE$Bootstrap_H1["DTEY",],95/100))
	  )

	tmp2 <- rbind(
	  c(TE$TransferEntropie_X - mean(TE$Bootstrap_H0["DTEX",])- 2*sd(TE$Bootstrap_H1["DTEX",]),TE$TransferEntropie_Y - mean(TE$Bootstrap_H0["DTEY",])- 2*sd(TE$Bootstrap_H1["DTEY",])),
	  c(TE$TransferEntropie_X - mean(TE$Bootstrap_H0["DTEX",]),TE$TransferEntropie_Y - mean(TE$Bootstrap_H0["DTEY",])),
	  c(TE$TransferEntropie_X - mean(TE$Bootstrap_H0["DTEX",])+ 2*sd(TE$Bootstrap_H1["DTEX",]),TE$TransferEntropie_Y - mean(TE$Bootstrap_H0["DTEY",]) + 2*sd(TE$Bootstrap_H1["DTEY",]))
	  )


	if (i==1){
	out <- c(ind[i],lags[lags[,2]==as.character(ind[i]),1],tex,stex,x_mboot,qx[1],qx[2],qx[3],sex,tey,stey,y_mboot,qy[1],qy[2],qy[3],sey,ndi,se_ndi,m_tex,m_tey,x_h1_bootmean,qx_h1[1],qx_h1[2],sdx_h1,y_h1_bootmean,qy_h1[1],qy_h1[2],sdy_h1)
	m1 <- tmp1
	m2 <- tmp2
	}else{
	out <- rbind(out,c(ind[i],lags[lags[,2]==as.character(ind[i]),1],tex,stex,x_mboot,qx[1],qx[2],qx[3],sex,tey,stey,y_mboot,qy[1],qy[2],qy[3],sey,ndi,se_ndi,m_tex,m_tey,x_h1_bootmean,qx_h1[1],qx_h1[2],sdx_h1,y_h1_bootmean,qy_h1[1],qy_h1[2],sdy_h1))
	m1 <- cbind(m1,tmp1)
	m2 <- cbind(m2,tmp2)
	}
      #   print(paste(ind[i],qx[1],stex,qx[2],sex,qy[1],stey,qy[2],sey,"\\",sep = " & "))
      #   print(paste(ind[i],lags[i],qx[1],stex,qx[2],sex,qy[1],stey,qy[2],sey,"\\",sep = " & "))
      #   print(paste(ind[i],tex,stex,tey,stey,sx,sy,sep = "  "))

      }

      colnames(out) <- c("Ticker","lags","TE_X","TE_X_Shuffle","TE_X_Boot","XQ90","XQ95","XQ99","SE_X","TE_Y","TE_Y_Shuffle","TE_Y_Boot","YQ90","YQ95","YQ99","SE_Y","NDI","SE_NDI","MeanBootX","MeanBootY","TE_X_H1_MeanBoot","XQ5_H1","XQ95_H1","SE_X_H1","TE_Y_H1_MeanBoot","YQ5_H1","YQ95_H1","SE_Y_H1")

      print(out)
      write.csv(out,file=paste(respath,dest_folder,"cds_quantile.csv",sep=""))







}


# define subfolder for results
dest_folder <- "full/"

# load table for lags 
lags <- read.table("lags.txt",header=FALSE)
forall(lags,ind,c(5,95),dest_folder)

dest_folder <- "before/"
lags <- read.table("lags_before.txt",header=FALSE)
forall(lags,ind,c(5,95),dest_folder,crisis="before")

dest_folder <- "crisis/"
lags <- read.table("lags_crisis.txt",header=FALSE)
forall(lags,ind,c(5,95),dest_folder,crisis="crisis")

dest_folder <- "after/"
lags <- read.table("lags_after.txt",header=FALSE)
forall(lags,ind,c(5,95),dest_folder,crisis="after")




print(Sys.time())

quit(save="no")