# 	***********************************************************	
# 	Using transfer entropy to measure information flows 
# 	between financial markets
#	Studies in Nonlinear Dynamics and Econometrics
#	-----------------------------------------------------------
# 	Funktionen zur Berechnung der Transfer Entropie
#	Thomas Dimpfl, Luis Huergo, Franziska J. Peter
#	last modified 13 April 2012     
# 	***********************************************************



# Kodierung der Stichprobe
#---------------------------

KodStp <- function(Stich,
                   OGUG=FALSE,
                   Quant=c(5,20,80,95),
                   Klassen=NULL,
                   Skal=1000){

    Pool   <- LETTERS
    KStich <- Stich

    if(OGUG){

        OG <- max(Stich)
        UG <- min(Stich)
    
        OSeq               <- UG+((OG-UG)/Klassen)*(0:(Klassen))
        OSeq[length(OSeq)] <- OG
    
#         print(paste("Kodierung der Stichprobe: ",Klassen," im Intervall ",UG,":",OG,sep=""))
    
        for (j in 1:(length(OSeq)-1)) {
    
                KStich  <- ifelse((((KStich >= OSeq[j])&(KStich < OSeq[j+1]))*(j!=(length(OSeq)-1))
                +((KStich >= OSeq[j])&(KStich <= OSeq[j+1]))*(j==(length(OSeq)-1))),j*Skal,KStich)
        }

    
    }else{
    
        Qtl        <- tapply(Quant, 
                             1:length(Quant), 
                             function(Quant, Stich) quantile(Stich, type=8, probs=Quant/100), 
                             Stich)
        Qtl        <- c(Qtl,max(Stich))
        Qlength    <- length(Qtl)
        Klassen    <- Qlength

# 	print(paste("Kodierung der Stichprobe: ",Quant, "% Quantil",sep=""))
        
        names(Qtl) <- c(Quant,100)
    
        for (j in 1:Qlength){

            KStich <- ifelse(KStich <=Qtl[j], j*Skal, KStich)
        }
        
    }
    
    KStich   <- KStich/Skal
    KodStich <- Pool[KStich]
    
 

    return(KodStich)

}


# Generierung der Clusters
#--------------------------

KlustGen <- function(X,
                     Y=NULL,
                     lX,
                     lY=NULL,
                     Prog = TRUE){

    n      <- length(X)

    dClust <- lX + 1

    if(!is.null(lY) && lY>lX) {

        dClust  <- lY + 1
        difflag <- lY - lX

    }

    nClust <- (n-dClust) + 1 #Vorsicht, hier habe ich etwas ge�ndert!


    #if(n/dClust==floor(n/dClust)) nClust <- nClust + 1

    ClustList <-  vector("list", nClust)

    if(is.null(Y)){

         if(Prog){

            for(i in 1:nClust){

                ClustList[[i]] <- X[i:(i+lX)]

            }

        }else{

            for(i in 1:nClust){

                ClustList[[i]] <- X[i:(i+lX-1)]

            }

        }

    }else{
        if(lX>=lY){

            if(Prog){

                for(i in 1:nClust){

                    ClustList[[i]] <- c(X[i:(i+lX)],Y[(i+lX-lY):(i+lX-1)])

                }

            }else{

                for(i in 1:nClust){

                    ClustList[[i]] <- c(X[i:(i+lX-1)],Y[(i+lX-lY):(i+lX-1)])

                }



            }

        }else{

            if(Prog){

                for(i in 1:nClust){

                    ClustList[[i]] <- c(X[(i + difflag):(i + difflag+ lX)],Y[i:(i+lY-1)])

                }

            }else{

                for(i in 1:nClust){

                    ClustList[[i]] <- c(X[(i + difflag):(i + difflag + lX -1)],Y[i:(i+lY-1)])

                }



            }

        }

    }

    KompDarst      <- unlist(lapply(ClustList, function(x) paste(x, collapse="")))

    MultWahr       <- table(KompDarst)/n


    return(list(Zeichenketten=KompDarst,
                Haeufigkeiten=MultWahr ))

}

TransEnt <- function(X,Y,lX,lY){

    # X = Lead! also TE Y -> X
    
    #Berechnung der Haeufigkeiten
    #------------------------------

    #X(k+1) und Y(j):
    #----------------

    Kplus1_J <- KlustGen(X,Y, lX=lX,lY=lY)$Haeufigkeiten

    AnzKlassKplus1_J <- length(Kplus1_J)

    #X(k+1):
    #-------

    Kplus1   <- KlustGen(X, lX=lX)$Haeufigkeiten

    AnzKlassKplus1   <- length(Kplus1)

    #X(k) und Y(j):
    #--------------

    K_J <- KlustGen(X,Y,lX=lX,lY=lY, Prog=F)$Haeufigkeiten

    AnzKlassK_J <- length(K_J)

    #X(k):
    #------

    K <- KlustGen(X, lX=lX, Prog=F)$Haeufigkeiten

    AnzKlassK   <- length(K)


    # Zusammenstellung der Transfer-Entropie:
    #-----------------------------------------

    EntVec <- numeric(length(Kplus1_J))

    for(i in 1:length(Kplus1_J)){

        Term1     <- Kplus1[substr(names(Kplus1_J[i]),1,(lX+1))]

        Term2     <- K_J[paste(unlist(strsplit(names(Kplus1_J[i]),split=NULL))[-(lX+1)], collapse="")]

        Term3     <- K[substr(names(Kplus1_J[i]),1,lX)]


    #changed TD: log10 hatte Luis, log2 hatte Franzi

        EntVec[i] <- Kplus1_J[i]*log2((Kplus1_J[i]*Term3)/(Term2*Term1))


    }

    return(list(TransEnt       =sum(EntVec),
                nKlassKplus1_J =AnzKlassKplus1_J,
                nKlassKplus1   =AnzKlassKplus1,
                nKlassK_J      =AnzKlassK_J,
                nKlassK        =AnzKlassK))
}


Shuffled_TE <- function(nrep=10, 
                        Diff=TRUE,
                        X,
                        lX, 
                        Y,
                        lY){
                        
    # X = Lead! also TE Y -> X

    
    
    n        <- length(X)

    nreps    <- round((nrep+1)/6)
    
    Shuffle1 <- parallel(replicate(nreps,TransEnt(X=X,Y=sample(Y, n, replace =TRUE),
		lX=lX,lY=lY)$TransEnt), mc.set.seed = TRUE)
    Shuffle2 <- parallel(replicate(nreps,TransEnt(X=X,Y=sample(Y, n, replace =TRUE),
		lX=lX,lY=lY)$TransEnt), mc.set.seed = TRUE)
    Shuffle3 <- parallel(replicate(nreps,TransEnt(X=X,Y=sample(Y, n, replace =TRUE), 
		lX=lX,lY=lY)$TransEnt), mc.set.seed = TRUE)
    Shuffle4 <- parallel(replicate(nreps,TransEnt(X=X,Y=sample(Y, n, replace =TRUE),
		lX=lX,lY=lY)$TransEnt), mc.set.seed = TRUE)
    Shuffle5 <- parallel(replicate(nreps,TransEnt(X=X,Y=sample(Y, n, replace =TRUE), 
		lX=lX,lY=lY)$TransEnt), mc.set.seed = TRUE)
    Shuffle6 <- parallel(replicate(nreps,TransEnt(X=X,Y=sample(Y, n, replace =TRUE),
		lX=lX,lY=lY)$TransEnt), mc.set.seed = TRUE)
    Shuffle  <- collect(list(Shuffle1,Shuffle2,Shuffle3,Shuffle4,Shuffle5,Shuffle6))

    Shuffled <- cbind((Shuffle[[1]]),(Shuffle[[2]]),(Shuffle[[3]]),(Shuffle[[4]]),(Shuffle[[5]]),(Shuffle[[6]]))

    
    if(Diff){
    
        TE             <- TransEnt(X=X,Y=Y,lX=lX,lY=lY)$TransEnt

        STED           <- TE- mean(Shuffled)
    
        return(STED)
        
    }else{
    
        STE             <- mean(Shuffled)
    
        return(STE)
    }
    
    

}


#================Bootstrap-Funktionen=========================#

# Zusammenlegung der Stichprobe
#-------------------------------


KollapseStp <- function(X,Y){

    Pool <- LETTERS

    n       <- length(X)
    NeuProz <- cbind(X,Y)
    NeuProz <- apply(NeuProz, 1, function(X)paste(X, collapse=""))
    
    Auspr       <- sort(unique(NeuProz))
    nAuspr      <- length(Auspr)
    
    AusprTab        <- Pool[1:nAuspr]
    names(AusprTab) <- Auspr
    
    KollStich   <- numeric(n)
    KollStich[] <- NA
    
    if(length(Pool)<nAuspr){
    
        print("Die Anzahl an Buchstaben ist nicht ausreichend")
        stop
    
    }else{

        for (j in 1:nAuspr){
    
                KollStich  <- ifelse(NeuProz == Auspr[j],Pool[j], KollStich)
        }
        
    }
    out <- list(AusprTab   = AusprTab,
                KollStich = KollStich)
    return(out)

}



# Dekodierung der Stichprobe
#------------------------------


DecodeStp <- function(Xneu,
                      AusprTab){

    n              <- length(Xneu)

    
    nAuspr         <- length(AusprTab)

    Index          <- 1:n

    ProzMat        <- matrix(NA, ncol=2, nrow=n)
    colnames(ProzMat) <- c("X","Y")

    for(i in 1:nAuspr){


        getPos <- Index[Xneu==AusprTab[i]]

        if(length(getPos)!=0){

            ProzMat[getPos,1] <- substr(names(AusprTab[i]),1,1)
            ProzMat[getPos,2] <- substr(names(AusprTab[i]),2,2)


        }
        
    }
    return(ProzMat)

}



#Berechnung der Wahrscheinlichkeiten
#------------------------------------

GenWahr_1 <- function(X,
                      lengthB){

    lX <- 1

    n  <- length(X)


    #1) Ohne Lags
    #--------------


    WKT_X <-  table(X)/n


    #2) Mit Lags
    #--------------


    dClust <- lX + 1

    nClust <- n-dClust+1


    #if(n/dClust==floor(n/dClust)) nClust <- nClust + 1

    ClustList <-  vector("list", nClust)

    for(i in 1:nClust){

        ClustList[[i]] <- X[i:(i+lX)]

    }

    KompDarst      <- unlist(lapply(ClustList, function(x) paste(x, collapse="")))

    nKD            <- length(KompDarst)

    MultWahr       <- table(KompDarst)/n

    N               <- length(MultWahr)

    # Laenge der Liste
    #-----------------

    Auspr            <- names(MultWahr)

    WahrVec          <- numeric(N)

    names(WahrVec)   <- Auspr


    for(i in 1:N){

        Term1     <- MultWahr[i]

        Term2     <- WKT_X[substr(Auspr[i],1,1)]

        WahrVec[i] <- Term1/Term2

    }



    TransProb    <-  split(WahrVec,substr(Auspr,1,1))


    IndexSplit   <- split(1:(n-lengthB),X[1:(n-lengthB)]) #vorsicht! hier m�glicher Fehler (+ 1?)

    out          <- list(WKT_X      = WKT_X,
                         TransProb  = TransProb,
                         IndexSplit = IndexSplit)

    return(out)

}

GenWahr_2 <- function(X,
                      lX){

    # Wahrscheinlichkeiten fuer step-wise markov bootstrap
  
    n  <- length(X)


    #1) Ohne Lags
    #--------------


    WKT_X <-  table(X)/n


    #2) Mit Lags
    #--------------


    dClust <- lX + 1

    nClust <- n-dClust+1


    #if(n/dClust==floor(n/dClust)) nClust <- nClust + 1

    ClustList <-  vector("list", nClust)

    for(i in 1:nClust){

        ClustList[[i]] <- X[i:(i+lX)]

    }

    KompDarst      <- unlist(lapply(ClustList, function(x) paste(x, collapse="")))

    nKD            <- length(KompDarst)

    MultWahr       <- table(KompDarst)/n

    N               <- length(MultWahr)

    # L�nge der Liste
    #-----------------

    Auspr            <- names(MultWahr)

    WahrVec          <- numeric(N)

    names(WahrVec)   <- Auspr


    for(i in 1:N){

        Term1     <- MultWahr[i]

        Term2     <- WKT_X[substr(Auspr[i],1,1)]

        WahrVec[i] <- Term1/Term2

    }



    TransProb    <-  split(WahrVec,substr(Auspr,1,1))

    out          <- list(WKT_X      = WKT_X,
                         TransProb  = TransProb)

    return(out)

}

# Bootstrap
#-----------


Markov_Boot_block <- function(X,
                         lengthB,
                         WKT_X,
                         TransProb,
                         IndexSplit){

#     BLOCKBOOTSTRAP 

    n         <- length(X)
    nB        <- n/lengthB


    #Bootstrap
    #-----------

    BootList <- vector(nB, mode="list")

    #Erste Ziehung (Z�hlt nicht)

    ErstZieh <- sample(names(WKT_X),1,prob=WKT_X)

    Ziehung  <- sample(IndexSplit[[ErstZieh]],1)

    BootList[[1]] <- X[Ziehung:(Ziehung+lengthB-1)]

    # Ziehungen

    for(i in 2:nB){


        if(length(names(TransProb[[BootList[[i-1]][lengthB]]]))==1){

             Ziehung1      <- names(TransProb[[BootList[[i-1]][lengthB]]])

        }else{

            Ziehung1      <- sample(names(TransProb[[BootList[[i-1]][lengthB]]]),1,
                                prob=TransProb[[BootList[[i-1]][lengthB]]])
        }

        Ziehung2      <- sample(IndexSplit[[substr(Ziehung1,2,2)]],1)

        BootList[[i]] <- X[Ziehung2:(Ziehung2+lengthB-1)]


    }

    BootVec <- unlist(BootList)

    return(BootVec)

}


Markov_Boot_step <- function(X,
                         lX){

#     BOOTSTRAP THE MARKOV CHAIN STEPWISE 

    n <- length(X)+50


    #Bootstrap
    #-----------

    BootList <- vector(n, mode="list")

    # Erste Ziehung 

    Wkeiten <- GenWahr_2(X=X, lX=lX)

    ErstZieh <- sample(names(Wkeiten$WKT_X),1,prob=Wkeiten$WKT_X)

    BootList[[1]] <- ErstZieh

    # Ziehungen

    for(i in 2:(lX+1)){

	Wkeiten <- GenWahr_2(X=X, lX=i-1)
	TransProb  = Wkeiten$TransProb

	if(length(names(TransProb[[BootList[[i-1]]]]))==1){

             Ziehung1      <- names(TransProb[[BootList[[i-1]]]])

        }else{

            Ziehung1      <- sample(names(TransProb[[BootList[[i-1]]]]),1,
                                prob=TransProb[[BootList[[i-1]]]])
        }

        BootList[[i]] <- substr(Ziehung1,i,i)

    }

    TransProb <- GenWahr_2(X=X, lX=lX)$TransProb

    for(i in (lX+2):n){


        if(length(names(TransProb[[BootList[[i-1]]]]))==1){

             Ziehung1      <- names(TransProb[[BootList[[i-1]]]])

        }else{

            Ziehung1      <- sample(names(TransProb[[BootList[[i-1]]]]),1,
                                prob=TransProb[[BootList[[i-1]]]])
        }

        BootList[[i]] <- substr(Ziehung1,lX+1,lX+1)


    }

    BootVec <- unlist(BootList)
    BootVec <- BootVec[51:(length(X)+50)] 


    return(BootVec)

}


TransBoot <- function(X,
                      AusprTab,
                      Shuffle = TRUE,
                      lengthB,
                      WKT_X, 
                      TransProb,
                      IndexSplit,
                      lx,
                      ly,
                      Konst=FALSE,
                      KonstX=NULL,
                      KonstY=NULL,
                      nrep=10){

    # Bootstrap
    #------------                      
                        
    BootX <-  Markov_Boot_block(X=X,
                           lengthB,
                           WKT_X = WKT_X,
                           TransProb = TransProb,
                           IndexSplit = IndexSplit) 
    
                                    

    # Dekodierung
    #-------------


    DecMat <- DecodeStp(Xneu     = BootX,
                        AusprTab = AusprTab)


    #Transfer Entropie der gebootstrapten Prozesse
    #----------------------------------------------
    
    
    #-------------
    # Ohne Shuffle
    #-------------
    
    
    
    #Lead=X
    #---------

    DTEX <-  TransEnt(X=DecMat[,1],Y=DecMat[,2],lX=lx,lY=ly)$TransEnt  

    # Lead=Y
    #---------
    
    DTEY <-  TransEnt(X=DecMat[,2],Y=DecMat[,1],lX=ly,lY=lx)$TransEnt  



    #-------------
    # Shuffle
    #-------------
    
    if(Shuffle){
    
        if(Konst){
        
        # Lead=X
        #---------
        
        DTEX <-  TransEnt(X=DecMat[,1],Y=DecMat[,2],lX=lx,lY=ly)$TransEnt  - KonstX
        
        # Lead=Y
        #---------
        
        DTEY <-  TransEnt(X=DecMat[,2],Y=DecMat[,1],lX=ly,lY=lx)$TransEnt  - KonstY
        
        
        
        }else{
        
            # Lead=X
            #---------
            
            DTEX <- Shuffled_TE(nrep=nrep, 
                                Diff=TRUE,
                                X = DecMat[,1],
                                lX = lx, 
                                Y = DecMat[,2],
                                lY = ly)
            
            
            # Lead=Y
            #---------
            
            DTEY <- Shuffled_TE(nrep=nrep, 
                                Diff=TRUE,
                                X = DecMat[,2],
                                lX = ly, 
                                Y = DecMat[,1],
                                lY = lx)
                                    
        }
        
    }else{
    
        #-------------
        # Ohne Shuffle
        #-------------
    
    
        #Lead=X
        #---------
        
        DTEX <-  TransEnt(X=DecMat[,1],Y=DecMat[,2],lX=lx,lY=ly)$TransEnt  
        
        # Lead=Y
        #---------
        
        DTEY <-  TransEnt(X=DecMat[,2],Y=DecMat[,1],lX=ly,lY=lx)$TransEnt  

    }




    TransEntBoot         <- c(DTEX, DTEY)   #Zuest X und dann Y!
    names(TransEntBoot)  <- c("DTEX", "DTEY")   
    
    return(TransEntBoot)

}

TransBoot_H1 <- function(X,
                      AusprTab,
                      Shuffle = TRUE,
                      lx,
		      ly,
                      Konst=FALSE,
                      KonstX=NULL,
                      KonstY=NULL,
                      nrep=10){



    # Bootstrap
    #------------                      
                        
    BootX <-  Markov_Boot_step(X=X,
                           lX=lx) 
    
                                    

    # Dekodierung
    #-------------


    DecMat <- DecodeStp(Xneu     = BootX,
                        AusprTab = AusprTab)


    #Transfer Entropie der gebootstrapten Prozesse
    #----------------------------------------------
    
    
    #-------------
    # Ohne Shuffle
    #-------------
    
    
    
    #Lead=X
    #---------

    DTEX <-  TransEnt(X=DecMat[,1],Y=DecMat[,2],lX=lx,lY=ly)$TransEnt  

    # Lead=Y
    #---------
    
    DTEY <-  TransEnt(X=DecMat[,2],Y=DecMat[,1],lX=ly,lY=lx)$TransEnt  



    #-------------
    # Shuffle
    #-------------
    
    if(Shuffle){
    
        if(Konst){
        
        # Lead=X
        #---------
        
        DTEX <-  TransEnt(X=DecMat[,1],Y=DecMat[,2],lX=lx,lY=ly)$TransEnt  - KonstX
        
        # Lead=Y
        #---------
        
        DTEY <-  TransEnt(X=DecMat[,2],Y=DecMat[,1],lX=ly,lY=lx)$TransEnt  - KonstY
        
        
        
        }else{
        
            # Lead=X
            #---------
            
            DTEX <- Shuffled_TE(nrep=nrep, 
                                Diff=TRUE,
                                X = DecMat[,1],
                                lX = lx, 
                                Y = DecMat[,2],
                                lY = ly)
            
            
            # Lead=Y
            #---------
            
            DTEY <- Shuffled_TE(nrep=nrep, 
                                Diff=TRUE,
                                X = DecMat[,2],
                                lX = ly, 
                                Y = DecMat[,1],
                                lY = lx)
                                    
        }
        
    }else{
    
        #-------------
        # Ohne Shuffle
        #-------------
    
    
        #Lead=X
        #---------
        
        DTEX <-  TransEnt(X=DecMat[,1],Y=DecMat[,2],lX=lx,lY=ly)$TransEnt  
        
        # Lead=Y
        #---------
        
        DTEY <-  TransEnt(X=DecMat[,2],Y=DecMat[,1],lX=ly,lY=lx)$TransEnt  

    }




    TransEntBoot         <- c(DTEX, DTEY)   #Zuerst X und dann Y!
    names(TransEntBoot)  <- c("DTEX", "DTEY")   
    
    return(TransEntBoot)

}

TransBoot_H0 <- function(X,Y,
                      Shuffle = TRUE,
                      lx,
                      ly,
                      Konst=FALSE,
                      KonstX=NULL,
                      KonstY=NULL,
                      nrep=10){
    


    BootX <- Markov_Boot_step(X=X,
			      lX=lx)


    BootY <- Markov_Boot_step(X=Y,
			      lX=ly)

     if(Shuffle){
    
        if(Konst){   

	    # Shuffle identisch für alle Wiederholungen, 
	    DTEX <- TransEnt(X=BootX,Y=Y,lX=lx,lY=ly)$TransEnt - KonstX
	    DTEY <- TransEnt(X=BootY,Y=X,lX=ly,lY=lx)$TransEnt - KonstY



        }else{
  
	    # Shuffle neu in allen Wiederholungen
	    KonstX <- Shuffled_TE(nrep=nrep,Diff=TRUE, X = BootX, lX=lx, Y = Y, lY = ly)
            KonstY <- Shuffled_TE(nrep=nrep, Diff=TRUE,X = BootY, lX=ly, Y = X, lY = lx)

	    DTEX <- TransEnt(X=BootX,Y=Y,lX=lx,lY=ly)$TransEnt - KonstX
	    DTEY <- TransEnt(X=BootY,Y=X,lX=ly,lY=lx)$TransEnt - KonstY

        }
        
    }else{
    
    # ohne Shuffle
    DTEX <- TransEnt(X=BootX,Y=Y,lX=lx,lY=ly)$TransEnt
    DTEY <- TransEnt(X=BootY,Y=X,lX=ly,lY=lx)$TransEnt 

    }

   
    TransEntBoot         <- c(DTEX, DTEY)   #Zuest X und dann Y!
    names(TransEntBoot)  <- c("DTEX", "DTEY")   
    
    return(TransEntBoot)

}


#===============Ende Bootstrap-Funktionen=====================#






